from typing import Literal, Union

import rframe

from .base_schemas import VersionedXeDoc


class FaxConfig(VersionedXeDoc):
    """fax configuration values for WFSim"""

    _ALIAS = "fax_configs"

    class Config:
        smart_union = True

    field: str = rframe.Index()
    experiment: Literal["1t", "nt", "nt_design"] = rframe.Index(default="nt")
    detector: Literal["tpc", "muon_veto", "neutron_veto"] = rframe.Index(default="tpc")
    science_run: str = rframe.Index()
    version: str = rframe.Index(default="v1")

    value: Union[int, float, bool, str, list, dict]
    resource: str



import rframe
import datetime

from .base_pmt_data import BasePmtData
from ..constants import DETECTOR


class VoltageSetting(BasePmtData):
    _ALIAS = "pmt_voltage_settings"

    name: str = rframe.Index(min_length=4, max_length=60)
    detector: DETECTOR = rframe.Index()
    pmt: int = rframe.Index(ge=0)

    active: bool = True
    created_by: str
    comments: str
    created_date: datetime.datetime

    voltage: float



"""
The 'hotspot_veto_threshold' provides a per-run cut threshold of the 'hotspot_veto_cut',
which vetos the events with local single electron rate above the threshold in each run.
Reference: xenon:xenonnt:analysis:hot_spot_cut_summary
"""

from ..base_corrections import TimeIntervalCorrection

class HotspotVetoThreshold(TimeIntervalCorrection):
    _ALIAS = 'hotspot_veto_thresholds'
    value: float



import re
from typing import ClassVar

import rframe
from .._settings import settings


def camel_to_snake(name):
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    return re.sub("([a-z0-9])([A-Z])", r"\1_\2", name).lower()


class XeDoc(rframe.BaseSchema):
    """Base class for all xedocs schemas."""

    _ALIAS: ClassVar = ""
    _CATEGORY: ClassVar = "general"

    _XEDOCS = {}

    class Config:
        allow_population_by_field_name = True

    def __init_subclass__(cls) -> None:
        super().__init_subclass__()

        if "_ALIAS" not in cls.__dict__:
            cls._ALIAS = camel_to_snake(cls.__name__)

        if cls._ALIAS and cls._ALIAS not in cls._XEDOCS:
            cls._XEDOCS[cls._ALIAS] = cls

    @classmethod
    def default_datasource(cls):
        """This method is called when a query method is
        called and no datasource is passed.
        """
        import xedocs

        return xedocs.default_datasource_for(cls)

    @classmethod
    def help(cls):
        help_str = f"""
            Schema name:   {cls.__name__}
            Alias:         {cls._ALIAS}
            Index fields:  {list(cls.get_index_fields())}
            Column fields: {list(cls.get_column_fields())}
        """
        print(help_str)

    def __repr__(self):
        idx_str = ", ".join(
            [f"{attr}={getattr(self, attr)}" for attr in self.get_index_fields()]
        )
        values_str = ", ".join(
            [f"{attr}={getattr(self, attr)}" for attr in self.get_column_fields()]
        )

        header = f"Xenon {type(self).__name__} Document"
        repr_str = f"""
        {header}
        {len(header)*'-'}
 
        Category:      {self._CATEGORY}
        Alias:         {self._ALIAS}
        Index:         {idx_str}
        Values:        {values_str}
        """
        return repr_str


class VersionedXeDoc(XeDoc):
    _ALIAS = ""

    version: str = rframe.Index(max_length=20)



import rframe

from ..base_corrections import TimeIntervalCorrection


class GlobalVersion(TimeIntervalCorrection):
    _ALIAS = "global_versions"

    config_name: str = rframe.Index(max_length=80)

    url: str



"""

The pattern likelihood cut defined in SR0 was defined with a data-driven S2 map and a Gaussian 
smeared likelihood definition, using the Tensorflow framework. The data-driven S2 map is trained 
using a neural network, using S2 optical Monte Carlo simulation as an initial guess, to best-fit 
event_area_per_channel. Details can be found in the SR0 note. As the SR0 method relied on trainable 
parameters for each PMT the S2 pattern likelihood needs to be retrained to reflect the changed PMT list in SR1.

SR0 S2 pattern map note
https://xe1t-wiki.lngs.infn.it/doku.php?id=shenyang:s2_pattern_likelihood (Shenyang)

SR1 S2 pattern map note
https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenon1t:jacques:nt_sr1_s2pl (Jacques)


"""

from ..base_references import BaseMap


class S2PatternMap(BaseMap):
    _ALIAS = "s2_pattern_maps"
    fmt = "binary"



"""
Correction: Region linear and circular for AB/CD partitions
Affects: Corrected areas

Two distinct patterns of evolution of single electron corrections between A+B and C+D. Distinguish thanks to linear and circular regions.
This TPC partitioning affects the single electron gains - both the average value and not - and the relative extraction efficiency.

SR0 wiki: https://xe1t-wiki.lngs.infn.it/doku.php?id=jlong:sr0_2_region_se_correction
SR1 wiki: https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xenonnt:noahhood:corrections:se_gain_ee_final
"""

import rframe

from ..base_corrections import TimeIntervalCorrection

class SingleElectronGainPartition(TimeIntervalCorrection):

    _ALIAS = "single_electron_gain_partition"
    region: str = rframe.Index(max_length=80)
    value: float




from typing import Any, Dict, Union

import rframe

from .base_schemas import VersionedXeDoc


class PluginLineage(VersionedXeDoc):
    _ALIAS = "plugin_lineages"

    class Config:
        smart_union = True

    data_type: str = rframe.Index()
    lineage_hash: str = rframe.Index()
    plugin_ALIAS: str = rframe.Index()

    config: Dict[str, Union[tuple, Any]]
    depends_on: Dict[str, str]

    @classmethod
    def from_context(cls, context, data_type, version="ONLINE", run_id="0"):
        key = context.key_for(run_id, data_type)
        lineage = dict(key.lineage)
        plugin, plugin_version, config = lineage.pop(data_type)
        depends_on = {
            dtype: context.key_for(run_id, dtype).lineage_hash for dtype in lineage
        }
        plugin_config = cls(
            version=version,
            data_type=data_type,
            lineage_hash=key.lineage_hash,
            plugin=plugin,
            plugin_version=plugin_version,
            config=config,
            depends_on=depends_on,
        )
        return plugin_config

    @classmethod
    def load_config(cls, datasource=None, **labels):
        configs = {}
        docs = cls.find(datasource, **labels)
        for doc in docs:
            configs[doc.lineage_hash] = dict(doc.config)
        for doc in docs:
            missing_deps = [
                lhash for lhash in doc.depends_on.values() if lhash not in configs
            ]
            if not missing_deps:
                continue
            for dep in cls.find(datasource, lineage_hash=missing_deps):
                configs[dep.lineage_hash] = dict(dep.config)
        combined = {}
        for config in configs.values():
            for k, v in config.items():
                if k in combined and v != combined[k]:
                    raise RuntimeError(
                        "Under-constrained search, " "results in inconsistent config."
                    )
                combined[k] = v
        return combined

    def get_dependencies(self, datasource=None, recursive=False):
        lineages = list(self.depends_on.values())
        docs = self.find(datasource, lineage_hash=lineages)
        deps = {doc.data_type for doc in docs}
        if recursive:
            for doc in docs:
                extra = doc.get_dependencies(datasource, recursive=True)
                deps.update(extra)
        return deps

    def get_full_config(self, datasource=None):
        config = {}
        for dep in self.get_dependencies(datasource):
            config.update(dep.config)
        return config



import datetime

import rframe
from pydantic import validator, BaseModel, constr
from typing import Literal, List

from ..base_schemas import XeDoc
from ..._settings import settings
from ..constants import SOURCE


class ActivityMeasurement(BaseModel):
    time: datetime.datetime
    activity: float
    uncertainty: float
    units: constr(max_length=10) = "Bq"


class CalibrationSource(XeDoc):
    _ALIAS = "calibration_sources"
    _CATEGORY = "calibration"

    source_id: str = rframe.Index(max_length=50)
    lngs_id: constr(max_length=30)
    kind: SOURCE
    ref: str
    comments: str

    activity_measurements: List[ActivityMeasurement]



import datetime
import time

import pytz
import pandas as pd


class SimpleClock:
    utc: bool
    tz_aware: bool
    cutoff_offset: float

    def __init__(self, utc=True, tz_aware=True, cutoff_offset=3600) -> None:
        self.utc = utc
        self.tz_aware = tz_aware
        self.cutoff_offset = cutoff_offset

    def current_datetime(self):
        if self.utc:
            now = datetime.datetime.utcnow()
        else:
            now = datetime.datetime.now()
        return self.normalize_tz(now)

    def cutoff_datetime(self, buffer=0.0):
        offset = datetime.timedelta(seconds=self.cutoff_offset + buffer)
        return self.current_datetime() + offset

    def normalize_tz(self, dt: datetime.datetime) -> datetime.datetime:
        if isinstance(dt, str):
            dt = pd.to_datetime(dt, utc=self.utc)
        if isinstance(dt, pd.Timestamp):
            dt = dt.to_pydatetime()
        if dt.tzinfo is not None:
            if dt.tzinfo.utcoffset(dt) is not None and self.utc:
                dt = dt.astimezone(pytz.utc)
        elif self.utc:
            dt = dt.replace(tzinfo=pytz.utc)
        if not self.tz_aware:
            dt = dt.replace(tzinfo=None)
            
        dt = dt.replace(microsecond=int(dt.microsecond / 1000) * 1000)
    
        return dt

    def after_cutoff(self, dt, buffer=0.0):
        cutoff = self.cutoff_datetime(buffer)
        dt = self.normalize_tz(dt)
        return dt > cutoff

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls, v):
        if not hasattr(v, "after_cutoff"):
            raise TypeError("Clock must implement `after_cutoff`")

        if not hasattr(v, "cutoff_datetime"):
            raise TypeError("Clock must implement `cutoff_datetime`")

        if not hasattr(v, "current_datetime"):
            raise TypeError("Clock must implement `current_datetime`")

        if not hasattr(v, "normalize_tz"):
            raise TypeError("Clock must implement `normalize_tz`")

        return v


class FrozenClock(SimpleClock):
    now: datetime.datetime

    def __init__(self, now: datetime.datetime, utc=True, cutoff_offset=3600) -> None:
        super().__init__(utc=utc, cutoff_offset=cutoff_offset)
        self.now = self.normalize_tz(now)

    def current_datetime(self):
        return self.now



from typing import Literal
from .base_report import BaseOperationsReport


class AnodeRampReport(BaseOperationsReport):
    """Anode ramp - report"""

    _ALIAS = "anode_ramps"

    direction: Literal["up", "down"]
    starting_voltage: float
    final_voltage: float
    voltage_unit: str = "volts"



import rframe
import datetime
from typing import Optional
from pydantic import Field


from .base_pmt_data import BasePmtData
from ..constants import DETECTOR


class PmtInstall(BasePmtData):
    _ALIAS = "pmt_installs"

    detector: DETECTOR = rframe.Index()
    pmt: int = rframe.Index(ge=0)
    array: str = rframe.Index()
    sector: str = rframe.Index()
    time: datetime.datetime = rframe.Index()

    serial_number: str = Field(max_length=60)

    position_x: float
    position_y: float
    position_z: Optional[float]

    signal_channel: Optional[int]
    signal_connector: Optional[int]
    signal_feedthrough: Optional[str]

    amplifier_crate: Optional[int]
    amplifier_fan: Optional[int]
    amplifier_plug: Optional[int]
    amplifier_serial: Optional[int]
    amplifier_slot: Optional[int]
    amplifier_channel: Optional[int]

    digitizer_channel: Optional[int]
    digitizer_crate: Optional[int]
    digitizer_module: Optional[int]
    digitizer_slot: Optional[int]

    high_voltage_crate: Optional[int]
    high_voltage_board: Optional[int]
    high_voltage_channel: Optional[int]
    high_voltage_connector: Optional[int]
    high_voltage_feedthrough: Optional[str]
    high_voltage_return: Optional[int]



"""
Peak Reconstruction Bias

Atul Prajapati, Carlo Fuselli, Chiara Di Donato

https://xe1t-wiki.lngs.infn.it/doku.php?id=xenon:xennont:peak_reconstruction_bias_sr2
Description: We know that there are multiple effects that introduce a bias in the reconstructed area. We estimate this bias with simulations produced with fuse. We compare the raw area that is the input, with the reconstrcuted area. We observe that effects like thresholds, afterpulses (AP) and noise introduce an energy dependent bias of the order of 2-3%. We want to provide a correction for the S1 and S2 areas to take these effects into account. Note that photo-ionisation also induces a similar bias, but this topic is not addressed in this notes (for now). 

"""
import rframe

from ..base_references import BaseResourceReference
from ...constants import SIGNAL

class PeakReconstructionBias(BaseResourceReference):
    _ALIAS = "peak_reconstruction_bias"
    fmt = "json"

    signal: SIGNAL = rframe.Index()

    value: str


"""
# Correction: Relative light yield
# What is it correction: Corrected areas (cS1)

# Description: Correction done to the area of S1s in our detector. The relative light yield corresponds to the number of photons in the visible light spectrum released by an interaction per energy. This correction helps us convert from the number of photons observed in an S1 event to an energy.

# Latest wiki reference: 

"""


import rframe

from ..base_corrections import TimeSampledCorrection


class RelativeLightYield(TimeSampledCorrection):

    _ALIAS = "relative_light_yield"
    value: float
